@@ -40,6 +40,26 @@ void usage(char **argv) {
 	exit(EXIT_FAILURE);
 }
 
+std::string read_json_file(const char *fname) {
+	std::string out;
+
+	FILE *f = fopen(fname, "r");
+	if (f == NULL) {
+		perror(optarg);
+		exit(EXIT_OPEN);
+	}
+
+	char buf[2000];
+	size_t nread;
+	while ((nread = fread(buf, sizeof(char), 2000, f)) != 0) {
+		out += std::string(buf, nread);
+	}
+
+	fclose(f);
+
+	return out;
+}
+
 int main(int argc, char **argv) {
 	int i;
 	const char *outtile = NULL;
@@ -60,6 +80,7 @@ int main(int argc, char **argv) {
 		{"output", required_argument, 0, 'o'},
 		{"filter-points-multiplier", no_argument, 0, 'm'},
 		{"feature-filter", required_argument, 0, 'j'},
+		{"feature-filter-file", required_argument, 0, 'J'},
 		{"preserve-input-order", no_argument, 0, 'o' & 0x1F},
 		{"accumulate-attribute", required_argument, 0, 'E'},
 		{"unidecode-data", required_argument, 0, 'u' & 0x1F},
@@ -72,6 +93,7 @@ int main(int argc, char **argv) {
 		{"no-tile-compression", no_argument, 0, 'd' & 0x1F},
 		{"clip-bounding-box", required_argument, 0, 'k' & 0x1F},
 		{"clip-polygon", required_argument, 0, 'l' & 0x1F},
+		{"clip-polygon-file", required_argument, 0, 'm' & 0x1F},
 
 		{0, 0, 0, 0},
 	};
@@ -122,6 +144,10 @@ int main(int argc, char **argv) {
 			filter = optarg;
 			break;
 
+		case 'J':
+			filter = read_json_file(optarg);
+			break;
+
 		case 'o' & 0x1F:
 			preserve_input_order = true;
 			break;
@@ -181,6 +207,12 @@ int main(int argc, char **argv) {
 			break;
 		}
 
+		case 'm' & 0x1F: {
+			clipbbox clip = parse_clip_poly(read_json_file(optarg));
+			clipbboxes.push_back(clip);
+			break;
+		}
+
 		default:
 			fprintf(stderr, "Unrecognized flag -%c\n", i);
 			usage(argv);
@@ -262,6 +294,7 @@ int main(int argc, char **argv) {
 	// clip the clip polygons, if any, to the tile bounds,
 	// to reduce their complexity
 
+	bool clipped_to_nothing = false;
 	if (clipbboxes.size() > 0) {
 		long long wx1 = (nx - buffer / 256.0) * (1LL << (32 - nz));
 		long long wy1 = (ny - buffer / 256.0) * (1LL << (32 - nz));
@@ -283,38 +316,47 @@ int main(int argc, char **argv) {
 
 			if (c.dv.size() > 0) {
 				c.dv = clip_poly_poly(c.dv, tile_bounds);
+
+				if (c.dv.size() == 0) {
+					clipped_to_nothing = true;
+					break;
+				}
 			}
 		}
 	}
 
-	json_object *json_filter = NULL;
-	if (filter.size() > 0) {
-		json_filter = parse_filter(filter.c_str());
-	}
+	std::string out;
 
-	for (auto const &s : sources) {
-		std::string tile;
-		char buf[1000];
-		int len;
-
-		FILE *f = fopen(s.tile.c_str(), "rb");
-		if (f == NULL) {
-			perror(s.tile.c_str());
-			exit(EXIT_FAILURE);
+	if (!clipped_to_nothing) {
+		json_object *json_filter = NULL;
+		if (filter.size() > 0) {
+			json_filter = parse_filter(filter.c_str());
 		}
 
-		while ((len = fread(buf, sizeof(char), 1000, f)) > 0) {
-			tile.append(std::string(buf, len));
+		for (auto const &s : sources) {
+			std::string tile;
+			char buf[1000];
+			int len;
+
+			FILE *f = fopen(s.tile.c_str(), "rb");
+			if (f == NULL) {
+				perror(s.tile.c_str());
+				exit(EXIT_FAILURE);
+			}
+
+			while ((len = fread(buf, sizeof(char), 1000, f)) > 0) {
+				tile.append(std::string(buf, len));
+			}
+			fclose(f);
+
+			input_tile t = s;
+			t.tile = std::move(tile);
+			its.push_back(std::move(t));
 		}
-		fclose(f);
 
-		input_tile t = s;
-		t.tile = std::move(tile);
-		its.push_back(std::move(t));
+		out = overzoom(its, nz, nx, ny, detail, buffer, keep, exclude, exclude_prefix, do_compress, NULL, demultiply, json_filter, preserve_input_order, attribute_accum, unidecode_data, simplification, tiny_polygon_size, bins, bin_by_id_list, accumulate_numeric, SIZE_MAX, clipbboxes);
 	}
 
-	std::string out = overzoom(its, nz, nx, ny, detail, buffer, keep, exclude, exclude_prefix, do_compress, NULL, demultiply, json_filter, preserve_input_order, attribute_accum, unidecode_data, simplification, tiny_polygon_size, bins, bin_by_id_list, accumulate_numeric, SIZE_MAX, clipbboxes);
-
 	FILE *f = fopen(outfile, "wb");
 	if (f == NULL) {
 		perror(outfile);
